# Generated by Django 2.2.19 on 2021-09-14 19:31

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('mellon', '0004_migrate_issuer'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='usersamlidentifier',
            unique_together={('issuer_fk', 'name_id')},
        ),
        migrations.RemoveField(
            model_name='usersamlidentifier',
            name='issuer',
        ),
        migrations.RenameField(
            model_name='usersamlidentifier',
            old_name='issuer_fk',
            new_name='issuer',
        ),
        migrations.AlterUniqueTogether(
            name='usersamlidentifier',
            unique_together={('issuer', 'name_id')},
        ),
    ]
