{% extends "mellon/base.html" %}
{% load i18n %}

{% block mellon_extra_scripts %}
  {% if error_redirect_after_timeout %}
    <meta http-equiv="refresh" content="{{ error_redirect_after_timeout }};url={{ next_url }}">
  {% endif %}
{% endblock %}

{% block mellon_content %}
<div id="mellon-authentication-failure" class="mellon-message">
  <h2 class="mellon-message-header">{% trans "Authentication failed" %}</h2>
  <p class="mellon-message-body">
    {% blocktrans %}The authentication has failed.{% endblocktrans %}
    {% if reason %}<p class="mellon-reason">{% trans "Reason" %}&nbsp;: {{ reason }}</p>{% endif %}
  </p>
  <p class="mellon-message-continue">
    <a class="mellon-link" href="{{ next_url }}">{% trans "Continue" %}</a>
  </p>
  {% if debug %}
  <!-- DEBUG INFO:
       Issuer: {{ issuer }}
       Message: {{ status_message }}
       Status codes: {{ status_codes|join:", " }}
       Relaystate: {{ relaystate }}
  -->
  {% endif %}
</div>
{% endblock %}
